"""Spade plugin to communicate over MQTT. Heavily inspired by aiomqtt."""

from typing import Any
import asyncio


import paho.mqtt.client as mqtt
from loguru import logger

from paho.mqtt.properties import Properties, PacketTypes
from paho.mqtt.reasoncodes import ReasonCode
from paho.mqtt.client import topic_matches_sub


class MqttQueues:
    """Object where the state of the mqtt client is stored."""

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self._msg_queues: dict[str, asyncio.Queue] = {}
        # Publish events
        self._pending_publish: dict[int, asyncio.Event] = {}
        # Connected event
        self._pending_connect: asyncio.Event | None = None
        # Disconnected event
        self._pending_disconnect: asyncio.Event | None = None
        # Subscribed events
        self._pending_subscribe: dict[int, asyncio.Event] = {}
        # Unsubscribed events
        self._pending_unsubscribe: dict[int, asyncio.Event] = {}
        self._loop = loop

    def topic_list(self):
        """Return the list of topic currently subscribed."""
        return self._msg_queues.keys()

    def connect_start(self) -> asyncio.Task:
        """Create an event to be set by the connect callback, and return a
        task that wait for it."""
        self._pending_connect = asyncio.Event()
        return asyncio.create_task(self._pending_connect.wait(), name="connected")

    def connect_done(self):
        """Called when the connection is done"""
        if self._pending_connect is not None:
            self._loop.call_soon_threadsafe(self._pending_connect.set)

    def has_pending_connect(self) -> bool:
        """Return true if a connection is pending"""
        return self._pending_connect is not None and not self._pending_connect.is_set()

    def disconnect_start(self) -> asyncio.Task:
        """Create an event to be set by the disconnect callback, and return a
        task that wait for it."""
        self._pending_disconnect = asyncio.Event()
        return asyncio.create_task(self._pending_disconnect.wait(), name="disconnected")

    def disconnect_done(self):
        """Called when the disconnection is done"""
        if self._pending_disconnect is not None:
            self._loop.call_soon_threadsafe(self._pending_disconnect.set)

    def has_pending_disconnect(self) -> bool:
        """Return true if a disconnection is pending"""
        return self._pending_disconnect is not None and not self._pending_disconnect.is_set()

    def publish_start(self, mid: int) -> asyncio.Task:
        """Create an event to be set when the publish is done and return a
        task that wait for it."""
        self._pending_publish[mid] = asyncio.Event()
        return asyncio.create_task(self._pending_publish[mid].wait(), name="published")

    def publish_done(self, key: int) -> None:
        """Set the event corresponding to a publish, and remove it from the
        pending publish"""
        if key in self._pending_publish:
            ev = self._pending_publish.pop(key)
            ev.set()

    def subcribe_start(self, mid: int) -> asyncio.Task:
        """Create a new event, and wait for it."""
        self._pending_subscribe[mid] = asyncio.Event()
        return asyncio.create_task(
            self._pending_subscribe[mid].wait(), name="subscribed"
        )

    def subscribe_done(self, key: int) -> None:
        """Set the event corresponding to this subscription, and remove it
        from the pending subscriptions."""
        if key in self._pending_subscribe:
            ev = self._pending_subscribe.pop(key)
            self._loop.call_soon_threadsafe(ev.set)
            ev.set()

    def unsubcribe_start(self, mid: int) -> asyncio.Task:
        """Create a new event, and wait for it."""
        self._pending_unsubscribe[mid] = asyncio.Event()
        return asyncio.create_task(
            self._pending_unsubscribe[mid].wait(), name="unsubscribed"
        )

    def unsubscribe_done(self, key: int) -> None:
        """Set the event corresponding to this subscription, and remove it from the
        pending subscriptions."""
        if key in self._pending_unsubscribe:
            ev = self._pending_unsubscribe.pop(key)
            self._loop.call_soon_threadsafe(ev.set)

    def put_msg(self, message):
        """Add a message in the queue"""
        found = False
        for sub in self._msg_queues:
            if topic_matches_sub(sub, message.topic):
                self._msg_queues[sub].put_nowait(message)
                found = True
        if not found:
            logger.warning(
                f"Warning: message on topic {message.topic} did not match any subscription"
            )

    async def get_msg(self, topic):
        """Wait for a message to arrive on a subscribed topic"""
        if topic in self._msg_queues:
            return await self._msg_queues[topic].get()
        else:
            logger.error(f"Must subscribe to topic {topic} first")

    def add_topic(self, topic):
        """Add a topic in the message queues"""
        self._msg_queues[topic] = asyncio.Queue()

    def del_topic(self, topic):
        """Remove a topic from the message queue."""
        if topic in self._msg_queues:
            del self._msg_queues[topic]


# MQTT callbacks


def on_mqtt_connect(
    client: mqtt.Client, userdata: MqttQueues, _flags, reason_code, _properties
):
    """Callback when a SUBACK is received"""
    if not userdata.has_pending_connect():
        return
    if reason_code.is_failure:
        logger.debug(f"Failed to connect: {reason_code}.")
    else:
        # we should always subscribe from on_connect callback to be
        # sure our subscribed is persisted across reconnections.
        userdata.connect_done()
        logger.debug("connected.")
        for topic in userdata.topic_list():
            client.subscribe(topic)


def on_mqtt_disconnect(
    _client: mqtt.Client,
    userdata: MqttQueues,
    _flags: mqtt.DisconnectFlags,
    _reason_code: ReasonCode,
    _properties: Properties | None = None,
) -> None:
    """Callback called when a UNSUBACK is received"""
    if not userdata.has_pending_disconnect():
        logger.warning("unexpected disconnection")
        return
    userdata.disconnect_done()


def on_mqtt_message(_client: mqtt.Client, userdata: MqttQueues, message):
    """Callback for message reception"""
    userdata.put_msg(message)


def on_mqtt_subscribe(
    _client: mqtt.Client,
    userdata: MqttQueues,
    mid: int,
    _reason_codes: list[ReasonCode],
    _properties: Properties | None = None,
) -> None:
    """Callback for subcriptions"""
    userdata.subscribe_done(mid)


def on_mqtt_unsubscribe(
    _client: mqtt.Client,
    userdata: MqttQueues,
    mid: int,
    _reason_codes: list[ReasonCode],
    _properties: Properties | None = None,
):
    """Callback called when a UNSUBACK is received"""
    userdata.unsubscribe_done(mid)


def on_mqtt_publish(
    _client: mqtt.Client,
    userdata: MqttQueues,
    mid: int,
    _reason_code: ReasonCode,
    _properties: Properties,
):
    """Get and trigger (set) the event."""
    userdata.publish_done(mid)


class MqttMixinError(BaseException):
    """Exceptions generated from the mqtt mixin"""

    def __init__(self, rc: int | None, *args: Any):
        super().__init__(*args)
        self.rc = rc

    def __str__(self):
        return f"[code:{self.rc.value}] {self.rc!s}"


class MqttMixin:
    """
    This mixin provides MQTT support to SPADE agents.
    It must be used as superclass of a spade.Agent subclass.
    """

    async def _hook_plugin_after_connection(self, *args, **kwargs):
        try:
            await super()._hook_plugin_after_connection(*args, **kwargs)
        except AttributeError:
            logger.debug("_hook_plugin_after_connection is undefined")
        # create a new mqtt client
        client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, protocol=mqtt.MQTTv5)
        # add an instance of the mqtt component to the agent
        self.mqtt = MqttComponent(client)

    async def _hook_plugin_before_connection(self, *args, **kwargs):
        """
        Overload this method to hook a plugin before connetion is done
        """
        try:
            await super()._hook_plugin_before_connection(*args, **kwargs)
        except AttributeError:
            logger.debug("_hook_plugin_before_connection is undefined")


class MqttComponent:
    """Mqtt client wrapper."""

    def __init__(self, client: mqtt.Client):
        self.clients = [client]
        self.queues = MqttQueues(asyncio.get_event_loop())
        self.init_client(client)

    def init_client(self, client):
        client.on_connect = on_mqtt_connect
        client.on_disconnect = on_mqtt_disconnect
        client.on_message = on_mqtt_message
        client.on_publish = on_mqtt_publish
        client.on_subscribe = on_mqtt_subscribe
        client.on_unsubscribe = on_mqtt_unsubscribe
        client.user_data_set(self.queues)

    def add_client(self):
        """create a new mqtt client and append it to the list of clients"""
        client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, protocol=mqtt.MQTTv5)
        self.clients.append(client)
        self.init_client(client)
        return client

    async def connect(
        self, broker, timeout, username=None, password=None, append=False, **kwargs
    ):
        """Connect to the MQTT broker (and start the loop). If another
        connection is done, create a new client and a new event loop"""
        if not append and self.clients[0].is_connected():
            raise RuntimeWarning("warning, client already connected")
        elif append:
            # try creating a new client and event loop
            client = self.add_client()
        else:
            client = self.clients[0]
        client.username = username
        client.password = password
        logger.debug(f" connecting to {broker}...")
        client.connect(broker, **kwargs)
        connected_task = self.queues.connect_start()
        client.loop_start()
        if self.queues.has_pending_connect():
            await asyncio.wait_for(connected_task, timeout=timeout)

    async def disconnect(self):
        """Disconnect from the MQTT broker (and stop the loop)"""
        for client in self.clients:
            client.disconnect()
            disconnected_task = self.queues.disconnect_start()
            await asyncio.wait_for(disconnected_task, timeout=5)
            client.loop_stop()

    def is_connected(self, client_id=0):
        """Check connection status"""
        return self.clients[client_id].is_connected()

    # SUBSCRIBER USE CASES
    async def subscribe(self, topic_name: str, client_id=0):
        """
        Subscribe to a topic.

        Args:
            topic: name of the topix to subscribe to
        """
        # add a queue for this topic
        self.queues.add_topic(topic_name)
        # if connected, subscribe immediately
        if self.clients[client_id].is_connected():
            logger.debug(f"subscribing to topic {topic_name}")
            result, mid = self.clients[client_id].subscribe(topic_name)
            if result != mqtt.MQTT_ERR_SUCCESS or mid is None:
                raise MqttMixinError(result, "Could not subscribe to topic")

            subscribed_task = self.queues.subcribe_start(mid)
            await asyncio.wait_for(subscribed_task, timeout=30)
        else:
            raise MqttMixinError("Could not subscribe to topic: not connected")

    async def receive(self, topic):
        """Asynchronously receive a message from a topic"""
        return await self.queues.get_msg(topic)

    async def unsubscribe(self, topic_name: str, client_id):
        """
        Unsubscribe from a topic.

        Args:
            topic_name (str): name of the mqtt topic.
        """
        if topic_name in self.queues.topic_list():
            logger.debug(f"unsubscribing from topic {topic_name}")
            result, mid = self.clients[client_id].unsubscribe(topic_name)
            if result != mqtt.MQTT_ERR_SUCCESS or mid is None:
                raise MqttMixinError("Could not unsubscribe from topic")
            unsubscribed_task = self.queues.unsubcribe_start(mid)
            await asyncio.wait_for(unsubscribed_task, timeout=5)
            self.queues.del_topic(topic_name)
        else:
            logger.error(f"Error unsubscribing from topic {topic_name}")

    # PUBLISHER USE CASES
    async def publish(
        self,
        topic: str,
        payload: str,
        qos=0,
        properties=None,
        retain=False,
        client_id=0,
    ):
        """
        Publish an item to a node.

        Args:
            topic (str): Name of the topic.

        """
        props = Properties(packetType=PacketTypes.PUBLISH)
        if properties:
            for k, v in properties.items():
                props.UserProperty = (k, v)

        info = self.clients[client_id].publish(
            topic=topic, payload=payload, qos=qos, retain=retain, properties=props
        )
        # Early out on error
        if info.rc != mqtt.MQTT_ERR_SUCCESS:
            raise MqttMixinError(info.rc, "Could not publish message")
        # Early out on immediate success
        if info.is_published():
            return
        pub_task = self.queues.publish_start(info.mid)
        await asyncio.wait_for(pub_task, timeout=5)



# use testcontainers to
# - start a MQTT broker (eclipse-mosquitto)
# - start a XMPP broker (spade run)
import time
import subprocess
import signal
import asyncio

import pytest

from datetime import datetime as dt, timedelta
from testcontainers.core.container import DockerContainer

import spade
from spade_mqtt.mqtt_plugin import MqttMixin


@pytest.fixture(scope="session")
def mosquitto_container():
    container = DockerContainer("eclipse-mosquitto").with_exposed_ports(1883)
    container.start()

    host = container.get_container_host_ip()
    port = container.get_exposed_port(1883)

    async def delayed_pub(delay, topic, payload):
        await asyncio.sleep(delay)
        container.exec(["mosquitto_pub", "-h", "localhost", "-t", topic, "-m", payload])

    yield host, port, delayed_pub

    container.stop()


@pytest.fixture(scope="session")
def spade_broker():
    proc = subprocess.Popen(["spade", "run", "--host", "127.0.0.1", "--memory"])
    time.sleep(2)
    yield "localhost", 5222

    proc.send_signal(signal.SIGINT)
    proc.wait()


def test_basic_agent_start(spade_broker):
    """Basic test to check that the fixture works as expected"""

    class TestBehaviour(spade.behaviour.OneShotBehaviour):
        async def run(self):
            assert self.agent
            self.agent.prop = True

    class TestAgent(spade.agent.Agent):
        prop = False
        test_bh: TestBehaviour

        async def setup(self) -> None:
            self.test_bh = TestBehaviour()
            self.add_behaviour(self.test_bh)

    async def start_ag1():
        """Launch one testing agent."""
        t1 = TestAgent("test1@localhost", "pwd")
        await t1.start(auto_register=True)  # use host and port
        assert t1.test_bh
        while not t1.test_bh.is_killed():
            try:
                await asyncio.sleep(1)
            except KeyboardInterrupt:
                break
        assert t1.prop

    host, port = spade_broker
    spade.run(start_ag1())


def test_sub_one(spade_broker, mosquitto_container):
    """Test 1: connect to one broker on a defined topic
    external client send messages on topic
    agent receive messages.
    another agent send "ping" message to check that
    agent undertest is right"""

    class PingReply(spade.behaviour.CyclicBehaviour):
        async def run(self):
            m = await self.receive(1)
            if m and m.body and "ping" in m.body:
                r = m.make_reply()
                r.body = "pong"
                await self.send(r)

    class Ping(spade.behaviour.PeriodicBehaviour):
        replied: list[bool] = []
        sent_date: list[dt] = []
        delay: list[timedelta|None] = []
        ping_num = 0

        def __init__(self, period, to):
            super().__init__(period=period)
            self.to = to

        async def run(self) -> None:
            assert self.agent
            thread_id = f"ping-{dt.now().timestamp()}"
            m = spade.message.Message(to=self.to, body="ping", thread=thread_id)
            template = spade.template.Template(thread=thread_id)
            await self.send(m)
            self.sent_date.append(dt.now())
            self.delay.append(None)
            self.replied.append(False)
            self.agent.add_behaviour(PingWaitReply(self, self.ping_num), template)
            self.ping_num += 1

        def record_reply(self, index):
            if index >= len(self.replied):
                raise RuntimeError("Got a reply for non-existant ping")
            self.replied[index] = True
            self.delay[index] = dt.now() - self.sent_date[index]

        def all_received(self):
            """Check if all ping requests have been replied"""
            return all(self.replied)

        def average_delay(self):
            """Get the average delay in microseconds"""
            valid_delays = [d.microseconds for d in self.delay if d]
            return sum(valid_delays) / len(valid_delays)

    class PingWaitReply(spade.behaviour.CyclicBehaviour):
        def __init__(self, parent: Ping, index: int):
            super().__init__()
            self.parent = parent
            self.index = index

        async def run(self):
            m = await self.receive(1)
            if m:
                assert m.body and m.body == "pong"
                self.parent.record_reply(self.index)

    class MQTTSub(spade.behaviour.CyclicBehaviour):
        subscribed = False
        received = []

        async def run(self) -> None:
            assert self.agent
            assert self.agent.mqtt
            if not self.subscribed:
                await self.agent.mqtt.subscribe("test/topic1")
                self.subscribed = True
            msg = await self.agent.mqtt.receive("test/topic1")
            self.received.append(msg)
            # Try to get only 2 messages
            if len(self.received) >= 2:
                self.kill()

    class PingerAgent(spade.agent.Agent):
        ping_bh: Ping

        async def setup(self) -> None:
            self.ping_bh = Ping(1, "test1@localhost")
            self.add_behaviour(self.ping_bh)

    class TestAgent(MqttMixin, spade.agent.Agent):
        test_bh: MQTTSub
        connect_done = False
        mqtt_port: int
        mqtt_host: str

        def __init__(self, jid, pwd, mqtt_host, mqtt_port):
            super().__init__(jid=jid, password=pwd)
            self.mqtt_host = mqtt_host
            self.mqtt_port = mqtt_port

        async def setup(self) -> None:
            self.add_behaviour(PingReply())
            await self.mqtt.connect(
                broker=self.mqtt_host, port=self.mqtt_port, timeout=1
            )
            assert self.mqtt.is_connected()
            self.connect_done = True
            self.test_bh = MQTTSub()
            self.add_behaviour(self.test_bh)

    async def start_ag1(mqtt_host, mqtt_port, pub_fn):
        """Launch one testing agent."""
        nonlocal success
        t1 = TestAgent("test1@localhost", "pwd", mqtt_host, mqtt_port)
        await t1.start(auto_register=True)  # use host and port
        # start pinger agent
        pinger = PingerAgent("pinger@localhost", "pwd")
        await pinger.start(auto_register=True)
        assert t1.test_bh
        assert t1.connect_done
        await pub_fn(5, "test/topic1", "hello1")
        await pub_fn(5, "test/topic1", "hello2")
        while not t1.test_bh.is_killed():
            try:
                await asyncio.sleep(1)
            except KeyboardInterrupt:
                break
        assert len(t1.test_bh.received) == 2
        assert t1.test_bh.received[0].payload == b"hello1"
        assert t1.test_bh.received[1].payload == b"hello2"
        assert all(pinger.ping_bh.replied)

        success = True

    host, port = spade_broker
    success = False
    spade.run(start_ag1(*mosquitto_container))
    assert success


# Test 2: same as one on a wildcard topic

# Test 3: same as one, but with two brokers

# Test 4: agent publish on a defined topic
# external tool gets the messages



# Test 5: publish while connected to two MQTT brokers



"""Demonstrate the MQTT plugin for spade.

A mqtt broker should be running, as well as a XMPP server.

"""

import asyncio
import spade
from spade_mqtt.mqtt_plugin import MqttMixin


class MQTTSubBehaviour(spade.behaviour.CyclicBehaviour):
    """Behaviour that subscribes to a topic and store the messages received."""

    def __init__(self, topic):
        spade.behaviour.CyclicBehaviour.__init__(self)
        self.topic = topic

    async def run(self):
        print(f"awaiting message on topic {self.topic}")
        msg = await self.agent.mqtt.receive(self.topic)
        if msg:
            print(f"got message with payload {msg.payload}")
            self.agent.last_msg = msg.payload.decode("utf-8")

    async def on_start(self):
        await self.agent.mqtt.subscribe(self.topic)

    async def on_end(self):
        await self.agent.mqtt.unsubscribe(self.topic)


class MQTTPubBehaviour(spade.behaviour.PeriodicBehaviour):
    """Increment a counter every second, and publish it to a mqtt topic."""

    def __init__(self, topic):
        spade.behaviour.PeriodicBehaviour.__init__(self, 1)
        self.topic = topic

    async def on_start(self):
        self.agent.t = 0

    async def run(self):
        self.agent.t += 1
        if self.agent.t >= 10:
            await self.agent.stop()
        m = f"tick {self.agent.t}"
        print(f"agent {self.agent.jid} publishing {m} on topic {self.topic}")
        await self.agent.mqtt.publish(
            self.topic, m, properties={"performative": "inform", "thread": "test"}, retain=True
        )


class PubSubAgent(MqttMixin, spade.agent.Agent):
    """Testing agent"""

    t: int = -1
    mqtt_behav: MQTTSubBehaviour = None
    last_msg: str = None

    async def setup(self):
        print(f"Agent {self.jid} starting...")
        try:
            await self.mqtt.connect("localhost", 30)
        except TimeoutError:
            print("Warning: could not connect to MQTT brocker localhost")
        self.mqtt_behav = MQTTSubBehaviour("test/topic1")
        self.add_behaviour(self.mqtt_behav)
        self.add_behaviour(MQTTPubBehaviour("test/topic1"))


def pub_agent():
    """Fixture that create an agent that periodically publish data"""

    class PubAg(MqttMixin, spade.agent.Agent):
        """Agent that publish to mqtt"""
        async def setup(self):
            self.mqtt.connect("localhost", 5)
            self.add_behaviour(MQTTPubBehaviour("test/topic1"))

    return PubAg("pub@localhost", "pwd")


async def start_ag1():
    """Launch one testing agent."""
    t1 = PubSubAgent("test1@localhost", "pwd")
    await t1.start(auto_register=True)
    # wait until behaviour is killed
    while not t1.mqtt_behav.is_killed():
        try:
            await asyncio.sleep(1)
        except KeyboardInterrupt:
            break
    await asyncio.sleep(1)
    await t1.mqtt.disconnect()
    await t1.stop()
    
    assert t1.t == 10, "publish behaviour did not run as expected"
    assert t1.last_msg == "tick 10", "subscribe behaviour did not run as expected"



def test_pubsub():
    """Test an agent that both publish and subscribe."""
    spade.run(start_ag1())

if __name__ == "__main__":
    spade.run(start_ag1())
